@@ -1,7 +1,7 @@
 /*
     sysfs.c - Part of libsensors, a library for reading Linux sensor data
     Copyright (c) 2005 Mark M. Hoffman <mhoffman@lightlink.com>
-    Copyright (C) 2007 Jean Delvare <khali@linux-fr.org>
+    Copyright (C) 2007-2008 Jean Delvare <khali@linux-fr.org>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -15,7 +15,8 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
+    MA 02110-1301 USA.
 */
 
 /* this define needed for strndup() */
@@ -28,35 +29,142 @@
 #include <stdlib.h>
 #include <limits.h>
 #include <errno.h>
-#include <sysfs/libsysfs.h>
+#include <dirent.h>
 #include "data.h"
 #include "error.h"
 #include "access.h"
 #include "general.h"
 #include "sysfs.h"
 
+
+/****************************************************************************/
+
+#define ATTR_MAX	128
+
+/*
+ * Read an attribute from sysfs
+ * Returns a pointer to a freshly allocated string; free it yourself.
+ * If the file doesn't exist or can't be read, NULL is returned.
+ */
+static char *sysfs_read_attr(const char *device, const char *attr)
+{
+	char path[NAME_MAX];
+	char buf[ATTR_MAX], *p;
+	FILE *f;
+
+	snprintf(path, NAME_MAX, "%s/%s", device, attr);
+
+	if (!(f = fopen(path, "r")))
+		return NULL;
+	p = fgets(buf, ATTR_MAX, f);
+	fclose(f);
+	if (!p)
+		return NULL;
+
+	/* Last byte is a '\n'; chop that off */
+	p = strndup(buf, strlen(buf) - 1);
+	if (!p)
+		sensors_fatal_error(__func__, "Out of memory");
+	return p;
+}
+
+/*
+ * Call an arbitrary function for each class device of the given class
+ * Returns 0 on success (all calls returned 0), a positive errno for
+ * local errors, or a negative error value if any call fails.
+ */
+static int sysfs_foreach_classdev(const char *class_name,
+				   int (*func)(const char *, const char *))
+{
+	char path[NAME_MAX];
+	int path_off, ret;
+	DIR *dir;
+	struct dirent *ent;
+
+	path_off = snprintf(path, NAME_MAX, "%s/class/%s",
+			    sensors_sysfs_mount, class_name);
+	if (!(dir = opendir(path)))
+		return errno;
+
+	ret = 0;
+	while (!ret && (ent = readdir(dir))) {
+		if (ent->d_name[0] == '.')	/* skip hidden entries */
+			continue;
+
+		snprintf(path + path_off, NAME_MAX - path_off, "/%s",
+			 ent->d_name);
+		ret = func(path, ent->d_name);
+	}
+
+	closedir(dir);
+	return ret;
+}
+
+/*
+ * Call an arbitrary function for each device of the given bus type
+ * Returns 0 on success (all calls returned 0), a positive errno for
+ * local errors, or a negative error value if any call fails.
+ */
+static int sysfs_foreach_busdev(const char *bus_type,
+				 int (*func)(const char *, const char *))
+{
+	char path[NAME_MAX];
+	int path_off, ret;
+	DIR *dir;
+	struct dirent *ent;
+
+	path_off = snprintf(path, NAME_MAX, "%s/bus/%s/devices",
+			    sensors_sysfs_mount, bus_type);
+	if (!(dir = opendir(path)))
+		return errno;
+
+	ret = 0;
+	while (!ret && (ent = readdir(dir))) {
+		if (ent->d_name[0] == '.')	/* skip hidden entries */
+			continue;
+
+		snprintf(path + path_off, NAME_MAX - path_off, "/%s",
+			 ent->d_name);
+		ret = func(path, ent->d_name);
+	}
+
+	closedir(dir);
+	return ret;
+}
+
+/****************************************************************************/
+
 char sensors_sysfs_mount[NAME_MAX];
 
 #define MAX_SENSORS_PER_TYPE	20
 #define MAX_SUBFEATURES		8
-/* Room for all 3 types (in, fan, temp) with all their subfeatures + VID
-   + misc features */
+#define MAX_SENSOR_TYPES	5
+/* Room for all 5 types (in, fan, temp, power, energy) with all their
+   subfeatures + VID + misc features */
 #define ALL_POSSIBLE_SUBFEATURES \
-				(MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES * 6 \
-				 + MAX_SENSORS_PER_TYPE + 1)
+				(MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES * \
+				 MAX_SENSOR_TYPES * 2 + \
+				 MAX_SENSORS_PER_TYPE + 1)
 
 static
 int get_type_scaling(sensors_subfeature_type type)
 {
+	/* Multipliers for subfeatures */
 	switch (type & 0xFF80) {
 	case SENSORS_SUBFEATURE_IN_INPUT:
 	case SENSORS_SUBFEATURE_TEMP_INPUT:
 		return 1000;
 	case SENSORS_SUBFEATURE_FAN_INPUT:
 		return 1;
+	case SENSORS_SUBFEATURE_POWER_AVERAGE:
+	case SENSORS_SUBFEATURE_ENERGY_INPUT:
+		return 1000000;
 	}
 
+	/* Multipliers for second class subfeatures
+	   that need their own multiplier */
 	switch (type) {
+	case SENSORS_SUBFEATURE_POWER_AVERAGE_INTERVAL:
 	case SENSORS_SUBFEATURE_VID:
 	case SENSORS_SUBFEATURE_TEMP_OFFSET:
 		return 1000;
@@ -74,6 +182,8 @@ char *get_feature_name(sensors_feature_type ftype, char *sfname)
 	case SENSORS_FEATURE_IN:
 	case SENSORS_FEATURE_FAN:
 	case SENSORS_FEATURE_TEMP:
+	case SENSORS_FEATURE_POWER:
+	case SENSORS_FEATURE_ENERGY:
 		underscore = strchr(sfname, '_');
 		name = strndup(sfname, underscore - sfname);
 		break;
@@ -133,6 +243,19 @@ static const struct subfeature_type_match fan_matches[] = {
 	{ NULL, 0 }
 };
 
+static const struct subfeature_type_match power_matches[] = {
+	{ "average", SENSORS_SUBFEATURE_POWER_AVERAGE },
+	{ "average_highest", SENSORS_SUBFEATURE_POWER_AVERAGE_HIGHEST },
+	{ "average_lowest", SENSORS_SUBFEATURE_POWER_AVERAGE_LOWEST },
+	{ "average_interval", SENSORS_SUBFEATURE_POWER_AVERAGE_INTERVAL },
+	{ NULL, 0 }
+};
+
+static const struct subfeature_type_match energy_matches[] = {
+	{ "input", SENSORS_SUBFEATURE_ENERGY_INPUT },
+	{ NULL, 0 }
+};
+
 static const struct subfeature_type_match cpu_matches[] = {
 	{ "vid", SENSORS_SUBFEATURE_VID },
 	{ NULL, 0 }
@@ -143,6 +266,8 @@ static struct feature_type_match matches[] = {
 	{ "in%d%c", in_matches },
 	{ "fan%d%c", fan_matches },
 	{ "cpu%d%c", cpu_matches },
+	{ "power%d%c", power_matches },
+	{ "energy%d%c", energy_matches },
 };
 
 /* Return the subfeature type and channel number based on the subfeature
@@ -176,22 +301,36 @@ sensors_subfeature_type sensors_subfeature_get_type(const char *name, int *nr)
 	return SENSORS_SUBFEATURE_UNKNOWN;
 }
 
+static int sensors_get_attr_mode(const char *device, const char *attr)
+{
+	char path[NAME_MAX];
+	struct stat st;
+	int mode = 0;
+
+	snprintf(path, NAME_MAX, "%s/%s", device, attr);
+	if (!stat(path, &st)) {
+		if (st.st_mode & S_IRUSR)
+			mode |= SENSORS_MODE_R;
+		if (st.st_mode & S_IWUSR)
+			mode |= SENSORS_MODE_W;
+	}
+	return mode;
+}
+
 static int sensors_read_dynamic_chip(sensors_chip_features *chip,
-				     struct sysfs_device *sysdir)
+				     const char *dev_path)
 {
 	int i, fnum = 0, sfnum = 0, prev_slot;
-	struct sysfs_attribute *attr;
-	struct dlist *attrs;
+	DIR *dir;
+	struct dirent *ent;
 	sensors_subfeature *all_subfeatures;
 	sensors_subfeature *dyn_subfeatures;
 	sensors_feature *dyn_features;
 	sensors_feature_type ftype;
 	sensors_subfeature_type sftype;
 
-	attrs = sysfs_get_device_attributes(sysdir);
-
-	if (attrs == NULL)
-		return -ENOENT;
+	if (!(dir = opendir(dev_path)))
+		return -errno;
 
 	/* We use a large sparse table at first to store all found
 	   subfeatures, so that we can store them sorted at type and index
@@ -199,29 +338,34 @@ static int sensors_read_dynamic_chip(sensors_chip_features *chip,
 	all_subfeatures = calloc(ALL_POSSIBLE_SUBFEATURES,
 				 sizeof(sensors_subfeature));
 	if (!all_subfeatures)
-		sensors_fatal_error(__FUNCTION__, "Out of memory");
+		sensors_fatal_error(__func__, "Out of memory");
 
-	dlist_for_each_data(attrs, attr, struct sysfs_attribute) {
-		char *name = attr->name;
+	while ((ent = readdir(dir))) {
+		char *name = ent->d_name;
 		int nr;
 
+		if (ent->d_name[0] == '.')
+			continue;
+
 		sftype = sensors_subfeature_get_type(name, &nr);
 		if (sftype == SENSORS_SUBFEATURE_UNKNOWN)
 			continue;
 
 		/* Adjust the channel number */
 		switch (sftype & 0xFF00) {
-			case SENSORS_SUBFEATURE_FAN_INPUT:
-			case SENSORS_SUBFEATURE_TEMP_INPUT:
-				nr--;
-				break;
+		case SENSORS_SUBFEATURE_FAN_INPUT:
+		case SENSORS_SUBFEATURE_TEMP_INPUT:
+		case SENSORS_SUBFEATURE_POWER_AVERAGE:
+		case SENSORS_SUBFEATURE_ENERGY_INPUT:
+			nr--;
+			break;
 		}
 
 		if (nr < 0 || nr >= MAX_SENSORS_PER_TYPE) {
 			/* More sensors of one type than MAX_SENSORS_PER_TYPE,
 			   we have to ignore it */
 #ifdef DEBUG
-			sensors_fatal_error(__FUNCTION__,
+			sensors_fatal_error(__func__,
 					    "Increase MAX_SENSORS_PER_TYPE!");
 #endif
 			continue;
@@ -231,11 +375,12 @@ static int sensors_read_dynamic_chip(sensors_chip_features *chip,
 		   sorted table */
 		switch (sftype) {
 		case SENSORS_SUBFEATURE_VID:
-			i = nr + MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES * 6;
+			i = nr + MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES *
+			    MAX_SENSOR_TYPES * 2;
 			break;
 		case SENSORS_SUBFEATURE_BEEP_ENABLE:
-			i = MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES * 6 +
-			    MAX_SENSORS_PER_TYPE;
+			i = MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES *
+			    MAX_SENSOR_TYPES * 2 + MAX_SENSORS_PER_TYPE;
 			break;
 		default:
 			i = (sftype >> 8) * MAX_SENSORS_PER_TYPE *
@@ -246,8 +391,7 @@ static int sensors_read_dynamic_chip(sensors_chip_features *chip,
 
 		if (all_subfeatures[i].name) {
 #ifdef DEBUG
-			sensors_fatal_error(__FUNCTION__,
-					    "Duplicate subfeature");
+			sensors_fatal_error(__func__, "Duplicate subfeature");
 #endif
 			continue;
 		}
@@ -257,13 +401,11 @@ static int sensors_read_dynamic_chip(sensors_chip_features *chip,
 		all_subfeatures[i].name = strdup(name);
 		if (!(sftype & 0x80))
 			all_subfeatures[i].flags |= SENSORS_COMPUTE_MAPPING;
-		if (attr->method & SYSFS_METHOD_SHOW)
-			all_subfeatures[i].flags |= SENSORS_MODE_R;
-		if (attr->method & SYSFS_METHOD_STORE)
-			all_subfeatures[i].flags |= SENSORS_MODE_W;
+		all_subfeatures[i].flags |= sensors_get_attr_mode(dev_path, name);
 
 		sfnum++;
 	}
+	closedir(dir);
 
 	if (!sfnum) { /* No subfeature */
 		chip->subfeature = NULL;
@@ -276,7 +418,8 @@ static int sensors_read_dynamic_chip(sensors_chip_features *chip,
 		if (!all_subfeatures[i].name)
 			continue;
 
-		if (i >= MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES * 6 ||
+		if (i >= MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES *
+		    MAX_SENSOR_TYPES * 2 ||
 		    i / (MAX_SUBFEATURES * 2) != prev_slot) {
 			fnum++;
 			prev_slot = i / (MAX_SUBFEATURES * 2);
@@ -286,7 +429,7 @@ static int sensors_read_dynamic_chip(sensors_chip_features *chip,
 	dyn_subfeatures = calloc(sfnum, sizeof(sensors_subfeature));
 	dyn_features = calloc(fnum, sizeof(sensors_feature));
 	if (!dyn_subfeatures || !dyn_features)
-		sensors_fatal_error(__FUNCTION__, "Out of memory");
+		sensors_fatal_error(__func__, "Out of memory");
 
 	/* Copy from the sparse array to the compact array */
 	sfnum = 0;
@@ -297,7 +440,8 @@ static int sensors_read_dynamic_chip(sensors_chip_features *chip,
 			continue;
 
 		/* New main feature? */
-		if (i >= MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES * 6 ||
+		if (i >= MAX_SENSORS_PER_TYPE * MAX_SUBFEATURES *
+		    MAX_SENSOR_TYPES * 2 ||
 		    i / (MAX_SUBFEATURES * 2) != prev_slot) {
 			ftype = all_subfeatures[i].type >> 8;
 			fnum++;
@@ -333,39 +477,68 @@ int sensors_init_sysfs(void)
 {
 	struct stat statbuf;
 
-	/* libsysfs will return success even if sysfs is not mounted,
-	   so we have to double-check */
-	if (sysfs_get_mnt_path(sensors_sysfs_mount, NAME_MAX)
-	 || stat(sensors_sysfs_mount, &statbuf) < 0
+	snprintf(sensors_sysfs_mount, NAME_MAX, "%s", "/sys");
+	if (stat(sensors_sysfs_mount, &statbuf) < 0
 	 || statbuf.st_nlink <= 2)	/* Empty directory */
 		return 0;
 
 	return 1;
 }
 
-/* returns: 0 if successful, !0 otherwise */
-static int sensors_read_one_sysfs_chip(struct sysfs_device *dev)
+/* returns: number of devices added (0 or 1) if successful, <0 otherwise */
+static int sensors_read_one_sysfs_chip(const char *dev_path,
+				       const char *dev_name,
+				       const char *hwmon_path)
 {
 	int domain, bus, slot, fn;
 	int err = -SENSORS_ERR_KERNEL;
-	struct sysfs_attribute *attr, *bus_attr;
-	char bus_path[SYSFS_PATH_MAX];
+	char *bus_attr;
+	char bus_path[NAME_MAX];
+	char linkpath[NAME_MAX];
+	char subsys_path[NAME_MAX], *subsys;
+	int sub_len;
 	sensors_chip_features entry;
 
 	/* ignore any device without name attribute */
-	if (!(attr = sysfs_get_device_attr(dev, "name")))
+	if (!(entry.chip.prefix = sysfs_read_attr(hwmon_path, "name")))
 		return 0;
 
-	/* NB: attr->value[attr->len-1] == '\n'; chop that off */
-	entry.chip.prefix = strndup(attr->value, attr->len - 1);
-	if (!entry.chip.prefix)
-		sensors_fatal_error(__FUNCTION__, "out of memory");
-
-	entry.chip.path = strdup(dev->path);
+	entry.chip.path = strdup(hwmon_path);
 	if (!entry.chip.path)
-		sensors_fatal_error(__FUNCTION__, "out of memory");
+		sensors_fatal_error(__func__, "Out of memory");
 
-	if (sscanf(dev->name, "%hd-%x", &entry.chip.bus.nr, &entry.chip.addr) == 2) {
+	if (dev_path == NULL) {
+		/* Virtual device */
+		entry.chip.bus.type = SENSORS_BUS_TYPE_VIRTUAL;
+		entry.chip.bus.nr = 0;
+		/* For now we assume that virtual devices are unique */
+		entry.chip.addr = 0;
+		goto done;
+	}
+
+	/* Find bus type */
+	snprintf(linkpath, NAME_MAX, "%s/subsystem", dev_path);
+	sub_len = readlink(linkpath, subsys_path, NAME_MAX - 1);
+	if (sub_len < 0 && errno == ENOENT) {
+		/* Fallback to "bus" link for kernels <= 2.6.17 */
+		snprintf(linkpath, NAME_MAX, "%s/bus", dev_path);
+		sub_len = readlink(linkpath, subsys_path, NAME_MAX - 1);
+	}
+	if (sub_len < 0) {
+		/* Older kernels (<= 2.6.11) have neither the subsystem
+		   symlink nor the bus symlink */
+		if (errno == ENOENT)
+			subsys = NULL;
+		else
+			goto exit_free;
+	} else {
+		subsys_path[sub_len] = '\0';
+		subsys = strrchr(subsys_path, '/') + 1;
+	}
+
+	if ((!subsys || !strcmp(subsys, "i2c")) &&
+	    sscanf(dev_name, "%hd-%x", &entry.chip.bus.nr,
+		   &entry.chip.addr) == 2) {
 		/* find out if legacy ISA or not */
 		if (entry.chip.bus.nr == 9191) {
 			entry.chip.bus.type = SENSORS_BUS_TYPE_ISA;
@@ -373,45 +546,47 @@ static int sensors_read_one_sysfs_chip(struct sysfs_device *dev)
 		} else {
 			entry.chip.bus.type = SENSORS_BUS_TYPE_I2C;
 			snprintf(bus_path, sizeof(bus_path),
-				"%s/class/i2c-adapter/i2c-%d/device/name",
+				"%s/class/i2c-adapter/i2c-%d/device",
 				sensors_sysfs_mount, entry.chip.bus.nr);
 
-			if ((bus_attr = sysfs_open_attribute(bus_path))) {
-				if (sysfs_read_attribute(bus_attr)) {
-					sysfs_close_attribute(bus_attr);
-					goto exit_free;
-				}
-
-				if (bus_attr->value
-				 && !strncmp(bus_attr->value, "ISA ", 4)) {
+			if ((bus_attr = sysfs_read_attr(bus_path, "name"))) {
+				if (!strncmp(bus_attr, "ISA ", 4)) {
 					entry.chip.bus.type = SENSORS_BUS_TYPE_ISA;
 					entry.chip.bus.nr = 0;
 				}
 
-				sysfs_close_attribute(bus_attr);
+				free(bus_attr);
 			}
 		}
-	} else if (sscanf(dev->name, "spi%hd.%d", &entry.chip.bus.nr,
-			  &entry.chip.addr) == 2) {
+	} else
+	if ((!subsys || !strcmp(subsys, "spi")) &&
+	    sscanf(dev_name, "spi%hd.%d", &entry.chip.bus.nr,
+		   &entry.chip.addr) == 2) {
 		/* SPI */
 		entry.chip.bus.type = SENSORS_BUS_TYPE_SPI;
-	} else if (sscanf(dev->name, "%*[a-z0-9_].%d", &entry.chip.addr) == 1) {
-		/* must be new ISA (platform driver) */
-		entry.chip.bus.type = SENSORS_BUS_TYPE_ISA;
-		entry.chip.bus.nr = 0;
-	} else if (sscanf(dev->name, "%x:%x:%x.%x", &domain, &bus, &slot, &fn) == 4) {
+	} else
+	if ((!subsys || !strcmp(subsys, "pci")) &&
+	    sscanf(dev_name, "%x:%x:%x.%x", &domain, &bus, &slot, &fn) == 4) {
 		/* PCI */
 		entry.chip.addr = (domain << 16) + (bus << 8) + (slot << 3) + fn;
 		entry.chip.bus.type = SENSORS_BUS_TYPE_PCI;
 		entry.chip.bus.nr = 0;
-	} else {
-		/* platform device with no id? */
+	} else
+	if ((!subsys || !strcmp(subsys, "platform") ||
+			!strcmp(subsys, "of_platform"))) {
+		/* must be new ISA (platform driver) */
+		if (sscanf(dev_name, "%*[a-z0-9_].%d", &entry.chip.addr) != 1)
+			entry.chip.addr = 0;
 		entry.chip.bus.type = SENSORS_BUS_TYPE_ISA;
 		entry.chip.bus.nr = 0;
-		entry.chip.addr = 0;
+	} else {
+		/* Ignore unknown device */
+		err = 0;
+		goto exit_free;
 	}
 
-	if (sensors_read_dynamic_chip(&entry, dev) < 0)
+done:
+	if (sensors_read_dynamic_chip(&entry, hwmon_path) < 0)
 		goto exit_free;
 	if (!entry.subfeature) { /* No subfeature, discard chip */
 		err = 0;
@@ -419,7 +594,7 @@ static int sensors_read_one_sysfs_chip(struct sysfs_device *dev)
 	}
 	sensors_add_proc_chips(&entry);
 
-	return 0;
+	return 1;
 
 exit_free:
 	free(entry.chip.prefix);
@@ -427,125 +602,105 @@ exit_free:
 	return err;
 }
 
+static int sensors_add_hwmon_device_compat(const char *path,
+					   const char *dev_name)
+{
+	int err;
+
+	err = sensors_read_one_sysfs_chip(path, dev_name, path);
+	if (err < 0)
+		return err;
+	return 0;
+}
+
 /* returns 0 if successful, !0 otherwise */
 static int sensors_read_sysfs_chips_compat(void)
 {
-	struct sysfs_bus *bus;
-	struct dlist *devs;
-	struct sysfs_device *dev;
-	int ret = 0;
+	int ret;
 
-	if (!(bus = sysfs_open_bus("i2c"))) {
-		if (errno && errno != ENOENT)
-			ret = -SENSORS_ERR_KERNEL;
-		goto exit0;
+	ret = sysfs_foreach_busdev("i2c", sensors_add_hwmon_device_compat);
+	if (ret && ret != ENOENT)
+		return -SENSORS_ERR_KERNEL;
+
+	return 0;
+}
+
+static int sensors_add_hwmon_device(const char *path, const char *classdev)
+{
+	char linkpath[NAME_MAX];
+	char device[NAME_MAX], *device_p;
+	int dev_len, err;
+	(void)classdev; /* hide warning */
+
+	snprintf(linkpath, NAME_MAX, "%s/device", path);
+	dev_len = readlink(linkpath, device, NAME_MAX - 1);
+	if (dev_len < 0) {
+		/* No device link? Treat as virtual */
+		err = sensors_read_one_sysfs_chip(NULL, NULL, path);
+	} else {
+		device[dev_len] = '\0';
+		device_p = strrchr(device, '/') + 1;
+
+		/* The attributes we want might be those of the hwmon class
+		   device, or those of the device itself. */
+		err = sensors_read_one_sysfs_chip(linkpath, device_p, path);
+		if (err == 0)
+			err = sensors_read_one_sysfs_chip(linkpath, device_p,
+							  linkpath);
 	}
-
-	if (!(devs = sysfs_get_bus_devices(bus))) {
-		if (errno && errno != ENOENT)
-			ret = -SENSORS_ERR_KERNEL;
-		goto exit1;
-	}
-
-	dlist_for_each_data(devs, dev, struct sysfs_device)
-		if ((ret = sensors_read_one_sysfs_chip(dev)))
-			goto exit1;
-
-exit1:
-	/* this frees bus and devs */
-	sysfs_close_bus(bus);
-
-exit0:
-	return ret;
+	if (err < 0)
+		return err;
+	return 0;
 }
 
 /* returns 0 if successful, !0 otherwise */
 int sensors_read_sysfs_chips(void)
 {
-	struct sysfs_class *cls;
-	struct dlist *clsdevs;
-	struct sysfs_class_device *clsdev;
-	int ret = 0;
+	int ret;
 
-	if (!(cls = sysfs_open_class("hwmon"))) {
+	ret = sysfs_foreach_classdev("hwmon", sensors_add_hwmon_device);
+	if (ret == ENOENT) {
 		/* compatibility function for kernel 2.6.n where n <= 13 */
 		return sensors_read_sysfs_chips_compat();
 	}
 
-	if (!(clsdevs = sysfs_get_class_devices(cls))) {
-		if (errno && errno != ENOENT)
-			ret = -SENSORS_ERR_KERNEL;
-		goto exit;
-	}
-
-	dlist_for_each_data(clsdevs, clsdev, struct sysfs_class_device) {
-		struct sysfs_device *dev;
-		if (!(dev = sysfs_get_classdev_device(clsdev))) {
-			ret = -SENSORS_ERR_KERNEL;
-			goto exit;
-		}
-		if ((ret = sensors_read_one_sysfs_chip(dev)))
-			goto exit;
-	}
-
-exit:
-	/* this frees cls and clsdevs */
-	sysfs_close_class(cls);
+	if (ret > 0)
+		ret = -SENSORS_ERR_KERNEL;
 	return ret;
 }
 
+/* returns 0 if successful, !0 otherwise */
+static int sensors_add_i2c_bus(const char *path, const char *classdev)
+{
+	sensors_bus entry;
+
+	if (sscanf(classdev, "i2c-%hd", &entry.bus.nr) != 1 ||
+	    entry.bus.nr == 9191) /* legacy ISA */
+		return 0;
+	entry.bus.type = SENSORS_BUS_TYPE_I2C;
+
+	/* Get the adapter name from the classdev "name" attribute
+	 * (Linux 2.6.20 and later). If it fails, fall back to
+	 * the device "name" attribute (for older kernels). */
+	entry.adapter = sysfs_read_attr(path, "name");
+	if (!entry.adapter)
+		entry.adapter = sysfs_read_attr(path, "device/name");
+	if (entry.adapter)
+		sensors_add_proc_bus(&entry);
+
+	return 0;
+}
+
 /* returns 0 if successful, !0 otherwise */
 int sensors_read_sysfs_bus(void)
 {
-	struct sysfs_class *cls;
-	struct dlist *clsdevs;
-	struct sysfs_class_device *clsdev;
-	sensors_bus entry;
-	int ret = 0;
+	int ret;
 
-	if (!(cls = sysfs_open_class("i2c-adapter"))) {
-		if (errno && errno != ENOENT)
-			ret = -SENSORS_ERR_KERNEL;
-		goto exit0;
-	}
+	ret = sysfs_foreach_classdev("i2c-adapter", sensors_add_i2c_bus);
+	if (ret && ret != ENOENT)
+		return -SENSORS_ERR_KERNEL;
 
-	if (!(clsdevs = sysfs_get_class_devices(cls))) {
-		if (errno && errno != ENOENT)
-			ret = -SENSORS_ERR_KERNEL;
-		goto exit1;
-	}
-
-	dlist_for_each_data(clsdevs, clsdev, struct sysfs_class_device) {
-		struct sysfs_device *dev;
-		struct sysfs_attribute *attr;
-
-		/* Get the adapter name from the classdev "name" attribute
-		 * (Linux 2.6.20 and later). If it fails, fall back to
-		 * the device "name" attribute (for older kernels). */
-		if (!(attr = sysfs_get_classdev_attr(clsdev, "name"))
-		 && !((dev = sysfs_get_classdev_device(clsdev)) &&
-		      (attr = sysfs_get_device_attr(dev, "name"))))
-			continue;
-
-		if (sscanf(clsdev->name, "i2c-%hd", &entry.bus.nr) != 1 ||
-		    entry.bus.nr == 9191) /* legacy ISA */
-			continue;
-		entry.bus.type = SENSORS_BUS_TYPE_I2C;
-
-		/* NB: attr->value[attr->len-1] == '\n'; chop that off */
-		entry.adapter = strndup(attr->value, attr->len - 1);
-		if (!entry.adapter)
-			sensors_fatal_error(__FUNCTION__, "out of memory");
-
-		sensors_add_proc_bus(&entry);
-	}
-
-exit1:
-	/* this frees *cls _and_ *clsdevs */
-	sysfs_close_class(cls);
-
-exit0:
-	return ret;
+	return 0;
 }
 
 int sensors_read_sysfs_attr(const sensors_chip_name *name,
